# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Text Embedding Module
=====================

Implements the following tasks:

  1. EmbeddingTask: Returns an embedding from an input text string
  2. EmbeddingsTasks: EmbeddingTask but with a list of inputs producing a list of outputs
  3. SentenceSimilarityTask: Compare one source sentence to a list of sentences
  4. SentenceSimilarityTasks: SentenceSimilarityTask but with a list of source sentences producing
     a list of outputs
  5. RerankTask: Return top_n documents ordered by relevance given a query
  6. RerankTasks: RerankTask but with a list of queries producing a list of outputs

"""

# Local
from .embedding import EmbeddingModule
